// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021-2022 Florian Fischer, Florian Schmaus
#include <fcntl.h>
#include <unistd.h>

#include <array>
#include <boost/program_options.hpp>
#include <climits>
#include <cstdlib>
#include <cstring>
#include <exception>
#include <filesystem>
#include <iostream>
#include <sstream>
#include <string>

#include "Common.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "Semaphore.hpp"
#include "StealingMode.hpp"
#include "io.hpp"

namespace fs = std::filesystem;
namespace po = boost::program_options;

#define EMPER_RIPGREP_BUFSIZE 4096

const char* needle;
size_t needle_len;

emper::Semaphore* max_running;

static enum emper::StealingMode stealingMode;

void search(const std::string& path) {
	if (max_running) {
		max_running->acquire();
	}

	int fd = emper::io::openAndWait(path.c_str(), O_RDONLY);
	if (fd < 0) {
		DIE_MSG_ERRNO("open of " << path << " failed");
	}

	std::array<char, EMPER_RIPGREP_BUFSIZE> buf;

	ssize_t bytes_read = emper::io::readFileAndWait(fd, buf.data(), buf.size(), 0);
	while (bytes_read > 0) {
		if (memmem(&buf[0], bytes_read, needle, needle_len)) {
			std::array<char, PATH_MAX + 1> outBuf;
			memcpy(outBuf.data(), path.c_str(), path.length());
			outBuf[path.length()] = '\n';
			emper::io::writeFileAndWait(STDOUT_FILENO, outBuf.data(), path.length() + 1, -1);
			goto out;
		}

		bytes_read = emper::io::readFileAndWait(fd, buf.data(), buf.size(), -1);
	}

	if (bytes_read < 0) {
		DIE_MSG_ERRNO("read of " << path << " failed");
	}

out:
	if (max_running) {
		max_running->release();
	}

	emper::io::closeAndForget(fd);
}

void walk_dir() {
	// Only search in regular files
	auto filter = [](const fs::directory_entry& dirent) { return dirent.is_regular_file(); };
	// Search the file
	auto fn = [](const fs::directory_entry& dirent) { search(dirent.path()); };

	switch (stealingMode) {
		case emper::StealingMode::child:
			emper::io::recursive_directory_walk(".", filter, fn);
			break;
		case emper::StealingMode::continuation:
			emper::io::fibril_recursive_directory_walk(".", filter, fn);
			break;
	}

	Runtime::getRuntime()->initiateTermination();
}

static auto fssearch(const po::variables_map& vm) -> int {
	if (!vm.count("needle")) {
		std::cerr << "No 'needle' specified" << std::endl;
		return EXIT_FAILURE;
	}

	auto needleStr = vm["needle"].as<std::string>();
	needle = needleStr.c_str();
	needle_len = needleStr.length();

	if (vm.count("max-fibers")) {
		unsigned int maxFibersCount = vm["max-fibers"].as<unsigned int>();
		max_running = new emper::Semaphore(maxFibersCount);
	}

	Runtime runtime;

	std::cerr << "Starting fsearch with stealingMode=" << stealingMode;
	if (max_running) std::cerr << " and " << max_running->getCount() << " file fibers";
	std::cerr << std::endl;

	auto* dirWalker = Fiber::from(walk_dir);
	runtime.scheduleFromAnywhere(*dirWalker);

	runtime.waitUntilFinished();

	delete max_running;

	return EXIT_SUCCESS;
}

auto main(int argc, char* argv[]) -> int {
	po::options_description desc("Allowed options");
	// clang-format off
	desc.add_options()
		("help", "Show help")
		("needle", po::value<std::string>(), "The String to search for")
		("max-fibers", po::value<unsigned int>(), "Maximum number of file processing fibers")
		("stealing-mode", po::value<enum emper::StealingMode>(&stealingMode)->default_value(emper::StealingMode::child), "Stealing mode to use, either 'child' or 'continuation'")
		;
	// clang-format on

	// Make 'needle' a positional option.
	po::positional_options_description pos_desc;
	pos_desc.add("needle", -1);

	// clang-format off
	auto parse_result = po::command_line_parser(argc, argv)
		.options(desc)
		.positional(pos_desc)
		.run()
		;
	// clang-format on

	po::variables_map vm;
	po::store(parse_result, vm);
	po::notify(vm);

	if (vm.count("help")) {
		std::cout << desc << "\n";
		return EXIT_SUCCESS;
	}

	try {
		return fssearch(vm);
	} catch (const std::exception& e) {
		std::cerr << e.what();
		return 1;
	}
}
